/* textview.c */

#include <glib.h>
#include <gdk/gdk.h>
#include <gdk/gdkkeysyms.h>
#include <gtk/gtkscrolledwindow.h>
#include <gtk/gtktext.h>
#include <gtk/gtksignal.h>
#include <stdio.h>
#include <ctype.h>
#include <string.h>

#include "main.h"
#include "headerview.h"
#include "summaryview.h"
#include "procheader.h"
#include "prefs_common.h"
#include "utils.h"
#include "unmime.h"
#include "procmime.h"

static GdkColor quote_color = {
	(gulong)0,
	(gushort)0,
	(gushort)0,
	(gushort)0xbfff
};

static GdkColor uri_color = {
	(gulong)0,
	(gushort)0,
	(gushort)0x8fff,
	(gushort)0
};

static GdkFont *msgfont;
static GdkFont *normalfont;
static GdkFont *boldfont;

static void textview_write_line		(TextView	*textview,
					 const gchar	*str,
					 gboolean	 charset_is_jis);
static MimeInfo *textview_scan_header	(TextView	*textview,
					 FILE		*fp);
static void mimeinfo_free		(MimeInfo	*mimeinfo);
static void textview_key_pressed	(GtkWidget	*widget,
					 GdkEventKey	*event,
					 TextView	*textview);

TextView *textview_create(void)
{
	TextView *textview;
	GtkWidget *scrolledwin;
	GtkWidget *text;

	debug_print(_("Creating text view...\n"));
	textview = g_new0(TextView, 1);

	scrolledwin = gtk_scrolled_window_new(NULL, NULL);
	gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(scrolledwin),
				       GTK_POLICY_NEVER, GTK_POLICY_ALWAYS);

	text = gtk_text_new(gtk_scrolled_window_get_hadjustment
			    (GTK_SCROLLED_WINDOW(scrolledwin)),
			    gtk_scrolled_window_get_vadjustment
			    (GTK_SCROLLED_WINDOW(scrolledwin)));
	gtk_widget_set_usize(scrolledwin, prefs_common.mainview_width, -1);
	gtk_container_add(GTK_CONTAINER(scrolledwin), text);
	gtk_signal_connect(GTK_OBJECT(text), "key_press_event",
			   GTK_SIGNAL_FUNC(textview_key_pressed),
			   textview);

	gtk_widget_show_all(scrolledwin);

	textview->scrolledwin = scrolledwin;
	textview->text        = text;

	return textview;
}

void textview_init(TextView *textview)
{
	GdkColormap *colormap;
	GdkColor color[2];
	gboolean success[2];

	color[0] = quote_color;
	color[1] = uri_color;
	colormap = gdk_window_get_colormap(textview->text->window);
	gdk_colormap_alloc_colors(colormap, color, 2, FALSE, TRUE, success);
	if (success[0] == FALSE || success[1] == FALSE) {
		GtkStyle *style;

		g_warning(_("TextView: color allocation failed\n"));
		style = gtk_widget_get_style(textview->text);
		quote_color = uri_color = style->black;
	}

	textview_set_font();
}

void textview_show_message(TextView *textview, const gchar *file)
{
	GtkText *text = GTK_TEXT(textview->text);
	gchar buf[BUFFSIZE], qp_buf[BUFFSIZE];
	MimeInfo *mimeinfo;
	FILE *fp;
	gboolean charset_is_jis = TRUE;
	gboolean is_qp = FALSE;
	gboolean softline = FALSE;

	if ((fp = fopen(file, "r")) == NULL) {
		perror("fopen");
		return;
	}

	gtk_text_freeze(text);
	gtk_text_backward_delete(text, gtk_text_get_length(text));

	/* if these aren't put, gtk+ will die by segmentation fault
	   (gtk+ bug?) */
	gtk_text_thaw(text);
	gtk_text_freeze(text);

	mimeinfo = textview_scan_header(textview, fp);

	if (mimeinfo) {
		//g_print("\nencoding: %s\nmime_type: %s\ncharset: %s\n",
		//	mimeinfo->encoding, mimeinfo->content_type,
		//	mimeinfo->charset);
		if (mimeinfo->charset) {
			if (!strcasecmp(mimeinfo->charset, "ISO-2022-JP"))
				charset_is_jis = TRUE;
			else
				charset_is_jis = FALSE;
		}
		if (mimeinfo->encoding_type == ENC_QUOTED_PRINTABLE)
			is_qp = TRUE;
	}

	if (!is_qp) {
		while (fgets(buf, sizeof(buf), fp) != NULL)
			textview_write_line(textview, buf, charset_is_jis);
	} else {
		while (fgets(buf, sizeof(buf), fp) != NULL || softline) {
			guchar *p = buf;

			if (softline) {
				softline = DoOneQPLine(&p, FALSE, softline);
				if (sizeof(qp_buf) - strlen(qp_buf)
				    >= strlen(buf) + 1) {
					strcat(qp_buf, buf);
					if (!softline)
						textview_write_line
							(textview, qp_buf,
							 charset_is_jis);
				} else {
					textview_write_line(textview, qp_buf,
							    charset_is_jis);
					if (!softline)
						textview_write_line
							(textview, buf,
							 charset_is_jis);
				}
			} else {
				softline = DoOneQPLine(&p, FALSE, softline);
				if (softline)
					strcpy(qp_buf, buf);
				else
					textview_write_line(textview, buf,
							    charset_is_jis);
			}

			buf[0] = '\0';
		}
	}

	procmime_mimeinfo_free(mimeinfo);
	fclose(fp);
	gtk_text_thaw(text);
}

void textview_write_line(TextView *textview, const gchar *str,
			 gboolean charset_is_jis)
{
	GtkText *text = GTK_TEXT(textview->text);
	gchar buf[BUFFSIZE];
	gchar *p;
	GdkColor *fg_color;

	if (charset_is_jis) {
		//conv_anytoeuc(buf, sizeof(buf), str);
		conv_jistoeuc(buf, sizeof(buf), str);
		conv_unreadable_eucjp(buf);
	} else {
		strcpy(buf, str);
		conv_unreadable_8bit(buf);
	}
	if (prefs_common.conv_mb_alnum) conv_mb_alnum(buf);
	fg_color = NULL;

	/* change color of quotation */
	if (prefs_common.enable_color &&
	    (p = strchr(buf, '>')) != NULL) {
		/* >, foo> ... ok, <foo>, foo bar> ... not ok */
		if (p - buf < 25 && memchr(buf, ' ', p - buf) == NULL &&
				    memchr(buf, '<', p - buf) == NULL)
			fg_color = &quote_color;
	}

	/* parse URI */
	if (prefs_common.enable_color &&
	    ((p = strcasestr(buf, "http://")) != NULL ||
	     (p = strcasestr(buf, "ftp://"))  != NULL)) {
		gchar *ep;

		gtk_text_insert(text, msgfont, fg_color, NULL, buf, p - buf);
		/* find end point of URI */
		for (ep = p; *ep; ep++)
			if (!isgraph(*ep) || !isascii(*ep) ||
			    strchr("()<>\"", *ep))
				break;
		gtk_text_insert(text, msgfont, &uri_color, NULL, p, ep - p);
		gtk_text_insert(text, msgfont, fg_color, NULL, ep, -1);
	} else
		gtk_text_insert(text, msgfont, fg_color, NULL, buf, -1);
}

void textview_quote_color_set(void)
{
}

void textview_set_font(void)
{
	if (prefs_common.messagefont)
		msgfont = gdk_fontset_load(prefs_common.messagefont);
	//if (!normalfont)
	//	normalfont = gdk_fontset_load(NORMAL_FONT);
	if (!boldfont)
		boldfont = gdk_fontset_load(BOLD_FONT);
}

enum
{
	H_DATE		= 0,
	H_FROM		= 1,
	H_TO		= 2,
	H_NEWSGROUPS	= 3,
	H_SUBJECT	= 4,
	H_CC		= 5,
	H_REPLY_TO	= 6,
	H_FOLLOWUP_TO	= 7,
	H_X_MAILER	= 8,
	H_X_NEWSREADER	= 9,
	H_USER_AGENT	= 10,
	H_ORGANIZATION	= 11,
	H_CONTENT_TRANSFER_ENCODING = 12,
	H_CONTENT_TYPE	= 13
};

static MimeInfo *textview_scan_header(TextView *textview, FILE *fp)
{
	static HeaderEntry hentry[] = {{"Date: ",         NULL, FALSE},
				       {"From: ",         NULL, TRUE},
				       {"To: ",	          NULL, FALSE},
				       {"Newsgroups: ",   NULL, FALSE},
				       {"Subject: ",      NULL, TRUE},
				       {"Cc: ",	          NULL, FALSE},
				       {"Reply-To: ",     NULL, FALSE},
				       {"Followup-To: ",  NULL, FALSE},
				       {"X-Mailer: ",     NULL, TRUE},
				       {"X-Newsreader: ", NULL, TRUE},
				       {"User-Agent: ",   NULL, TRUE},
				       {"Organization: ", NULL, TRUE},

				       {"Content-Transfer-Encoding: ",
							  NULL, FALSE},
				       {"Content-Type: ", NULL, FALSE},

				       {NULL,	          NULL, FALSE}};
	gchar buf[BUFFSIZE], tmp[BUFFSIZE];
	GtkText *text = GTK_TEXT(textview->text);
	gint hnum;
	HeaderEntry *hp;
	MsgInfo msginfo;
	MimeInfo *mimeinfo = NULL;

	g_return_val_if_fail(fp != NULL, NULL);

	memset(&msginfo, 0, sizeof(MsgInfo));

	/* show specified headers */
	while ((hnum = procheader_get_one_field(buf, sizeof(buf), fp, hentry))
	       != -1) {
		hp = hentry + hnum;

		if (H_CONTENT_TRANSFER_ENCODING == hnum) {
			if (!mimeinfo) mimeinfo = procmime_mimeinfo_new();
			procmime_scan_encoding
				(mimeinfo, buf + strlen(hp->name));
			continue;
		} else if (H_CONTENT_TYPE == hnum) {
			if (!mimeinfo) mimeinfo = procmime_mimeinfo_new();
			procmime_scan_content_type
				(mimeinfo, buf + strlen(hp->name));
			continue;
		}

		gtk_text_insert(text, boldfont, NULL, NULL, buf,
				strlen(hp->name));
		UnMimeHeader(buf);
		conv_jistoeuc(tmp, sizeof(tmp), buf + strlen(hp->name));
		conv_unreadable_eucjp(tmp);
		gtk_text_insert(text, normalfont, NULL, NULL, tmp, -1);
		gtk_text_insert(text, NULL, NULL, NULL, "\n", 1);

		switch (hnum) {
		case H_FROM:
			g_free(msginfo.from);
			msginfo.from = g_strdup(tmp);
			break;
		case H_TO:
			g_free(msginfo.to);
			remove_return(tmp);
			msginfo.to = g_strdup(tmp);
			break;
		case H_NEWSGROUPS:
			g_free(msginfo.newsgroups);
			remove_return(tmp);
			msginfo.newsgroups = g_strdup(tmp);
			break;
		case H_SUBJECT:
			g_free(msginfo.subject);
			msginfo.subject = g_strdup(tmp);
			break;
		}
	}

	gtk_text_insert(text, NULL, NULL, NULL, "\n", 1);

	headerview_show(textview->messageview->headerview, &msginfo);
	g_free(msginfo.from);
	g_free(msginfo.to);
	g_free(msginfo.newsgroups);
	g_free(msginfo.subject);

	return mimeinfo;
}

static void mimeinfo_free(MimeInfo *mimeinfo)
{
	if (!mimeinfo) return;

	g_free(mimeinfo->encoding);
	g_free(mimeinfo->content_type);
	g_free(mimeinfo->charset);
}

static void textview_key_pressed(GtkWidget *widget, GdkEventKey *event,
				 TextView *textview)
{
	SummaryView *summaryview =
		textview->messageview->mainwin->summaryview;

	if (!event) return;

	switch (event->keyval) {
	case GDK_Tab:
	case GDK_Home:
	case GDK_Left:
	case GDK_Up:
	case GDK_Right:
	case GDK_Down:
	case GDK_Page_Up:
	case GDK_Page_Down:
	case GDK_End:
	case GDK_Control_L:
	case GDK_Control_R:
		break;
	default:
		summary_pass_key_press_event(summaryview, event);
		gtk_widget_grab_focus(summaryview->ctree);
		break;
	}
}
